@@ -196,7 +196,11 @@ def get_callback(
 ):
     """Returns the Guide.process() callback."""
 
-    guider = Guider(config["observatory"], target_rms=target_rms)
+    guider = Guider(
+        config["observatory"],
+        target_rms=target_rms,
+        command=params.command,
+    )
     params.command.actor.state._guider_obj = guider  # To update PID coeffs.
 
     return partial(
@@ -210,6 +214,7 @@ def get_callback(
         gaia_cross_correlation_blur=params.cross_match_blur,
         fit_all_detections=params.fit_all_detections,
         plot=params.plot,
+        fit_focus=config["guider"].get("fit_focus", True),
         **params.mode_kwargs,
     )
 
