@@ -25,7 +25,7 @@ __all__ = ["reprocess"]
 @click.argument("MJD", type=int)
 @click.argument("FRAME", type=int)
 async def reprocess(command: ChernoCommandType, mjd: int, frame: int):
-    """Reprocesses a frame (does not apply corrections."""
+    """Reprocesses a frame (does not apply corrections)."""
 
     paths = list(pathlib.Path(f"/data/gcam/{mjd}/").glob(f"gimg-*{frame:04d}.fits"))
     images = [str(pp) for pp in paths if pp.exists()]
